/*  debug_malloc.c
 *	-- a debugging replacement for the "safe_malloc" package
 *  Thomas Padron-McCarthy (padrone@lysator.liu.se), 1991
 *  Latest change of this file: Oct 3, 1992
 *
 *  This module, "debug_malloc", is part of the safe_malloc/debug_malloc
 *  (SM/DM) package for malloc debugging.
 *
 *  This version of the malloc() functions performs some useful
 *  run-time checks.
 *  #define DEBUG_MALLOC (a good place is in CFLAGS in your Makefile)
 *  and recompile everything that uses any "safe_malloc" functions.
 *  Finally, link with "debug_malloc.o" instead of "safe_malloc.o".
 *
 */

#include <stdio.h>
#include <ctype.h>

#ifdef USE_STANDARD_MALLOC
#   include <malloc.h>
#endif

#include "smdm_config.h"

/* Don't #include "safe_malloc.h" in this file -- it re-defines malloc() etc! */

/*---------------------------------------------------------------------------*/

#ifdef MSDOS
	/* MSDOS mallocs sometimes allocate memory on even 64K boundaries */

#   ifdef USE_STANDARD_MALLOC
#       define HASH(a)		(((unsigned long)(a) >> 16) % HASH_TABLE_SIZE)
#   else
#       define HASH(a)		(((unsigned long)(a)) % HASH_TABLE_SIZE)
#   endif

#else
#   define HASH(a)		(((unsigned long)(a)) % HASH_TABLE_SIZE)
#endif

/*---------------------------------------------------------------------------*/

enum status_type { used, freed, realloced };

struct malloc_entry {
    long int entry_number;	/* The number of this entry */
    char *file_name;		/* The file name from which this area was allocated */
    int line_number;
    char *freed_file_name;	/* Also used for realloc */
    int freed_line_number;
    char *remark;		/* Any comment concerning this allocated area */
    char *the_area;
    /* With sixteen extra bytes, eight before and eight after!
       But "the_area" points to the real, used part of it! */
    SIZE_T the_size;
    /* This is the size without those extra sixteen bytes. */
    enum status_type status;
    /* Used, freed or realloced. */
    int checksum;
    struct malloc_entry *realloced_to;
    struct malloc_entry *realloced_from;
    struct malloc_entry *next_in_bucket, *next_in_sequence;
    /* We need "next_in_sequence" when going through the malloc entries
       in chronological order. They form a linked list. */
}; /* struct malloc_entry */

/*---------------------------------------------------------------------------*/

extern char *ctime();
extern int errno;
extern char *sys_errlist[];
extern int sys_nerr;

extern void dm_dump();
extern void dm_full_dump();

#ifndef USE_STANDARD_MALLOC
    static char *bottom_malloc();
#endif

/*---------------------------------------------------------------------------*/

static long int nr_entries = 0;	/* Total number of allocated areas (all are in use) */
static long int malloc_calls = 0, calloc_calls = 0, realloc_calls = 0, free_calls = 0;
				/* This is just the number of calls to the functions */
static int nr_warnings = 0;	/* Number of warnings given */

/* These variables can be changed from the application: */
FILE *dm_error_stream = stderr;
char *dm_dump_file_name = DEFAULT_DUMP_FILE_NAME;
int dm_insane = 0;
int dm_repair = 1;
int dm_reuse = 0;
int dm_fill_new = 1;
int dm_fill_freed = 1;
int dm_print_codes = 1;
char dm_print_char = DEFAULT_PRINT_CHARACTER;
int dm_max_print_length = DEFAULT_MAX_PRINT_LENGTH;
char *dm_global_remark = NULL;

char *dm_version = DEBUG_MALLOC_VERSION;

/*---------------------------------------------------------------------------*/

/*  Yes, I know. I should use varargs or something.
 *  But that's even LESS portable than this well-known kludge!
 */
static void warning(file, fun_name, file_name, line_nr, fmt, a1, a2, a3, a4, a5)
  FILE *file;
  char *fun_name, *file_name;
  int line_nr;
  char *fmt, *a1, *a2, *a3, *a4, *a5;
{
    ++nr_warnings;
    fprintf(file, "*** MALLOC WARNING %d", nr_warnings);
    if (fun_name == NULL)
	fprintf(file, ": ");
    else if (file_name == NULL)
	fprintf(file, " in %s called from an unknown place:\n",
		fun_name);
    else
	fprintf(file, " in %s at line number %d in '%s':\n    ",
		fun_name, line_nr, file_name);
    fprintf(file, fmt, a1, a2, a3, a4, a5);
    fprintf(file, "\n");
} /* warning */

static void fatal_error(fmt, a1, a2, a3, a4, a5)
  char *fmt, *a1, *a2, *a3, *a4, *a5;
{
    static int fatal_already = 0;

    if (!fatal_already) {
	fatal_already = 1;
	fprintf(stderr, "*** INTERNAL ERROR in the debug_malloc package:\n    ");
	fprintf(stderr, fmt, a1, a2, a3, a4, a5);
	fprintf(stderr, "\n");
	fprintf(stderr, "    Unix error number: %d (%s).\n",
		errno, sys_errlist[errno]);
	fprintf(stderr, "    Goodbye. I am sorry that this happened. Exiting.\n");
	fflush(stderr);
	dm_dump();
    }
    exit(1);
} /* fatal_error */

/*---------------------------------------------------------------------------*/

static char *real_malloc(size)
  SIZE_T size;
{
#ifdef USE_STANDARD_MALLOC
    return (char *)malloc(size);	/* A perfectly ordinary malloc call! */
#else
    return bottom_malloc(size);
#endif
} /* real_malloc */

static char *real_safe_malloc(size)
  SIZE_T size;
{
    char *temp;

    temp = real_malloc(size);

    if (temp == NULL) {
	fatal_error("Couldn't allocate space (%d bytes) for internal malloc data.", size);
	return NULL;
    }
    else
	return temp;
} /* real_safe_malloc */

/*---------------------------------------------------------------------------*/

/*  Lots of time here (says gprof), so I tried to make it more efficient.
 *  By changing from a for loop with an integer loop variable to a while
 *  loop with pointers, I once got the time spent in this function down
 *  from 1.09 to 0.40 for a certain application.
 */
static int calculate_checksum(cp, size)
  register char *cp;
  SIZE_T size;
{
    register int sum;
    register char *end_cp;

    sum = 0;
    end_cp = cp + size;
    while (cp != end_cp) {
	sum += *cp++;
	sum &= 0x00ff;		/* sum %= 256; */
    }
    return sum;
} /* calculate_checksum */

static void set_area(cp, size, what)
  register char *cp;
  SIZE_T size;
  register int what;
{
    register char *end_cp;

    end_cp = cp + size;
    while (cp != end_cp)
	*cp++ = what;
} /* set_area */

/* This simple copying function is not guaranteed to handle overlapping areas: */
static void copy_bytes(from, to, size)
  register char *from, *to;
  SIZE_T size;
{
    register char *end_cp;

    end_cp = from + size;
    while (from != end_cp)
	*to++ = *from++;
} /* copy_bytes */

/*---------------------------------------------------------------------------*/

/*  When calling "free" or "realloc", we must find the malloc entry
 *  associated with a certain address. We use a hash table for that.
 */

static struct malloc_entry *hash_table[ HASH_TABLE_SIZE ];
    /* Automatically initialized to all NULLs. */

/*  This function puts a malloc entry in its bucket.
 *  No two malloc entries will ever have the same key.
 */
static void hash_insert(mep)
  struct malloc_entry *mep;
{
    int index = HASH(mep->the_area);

    mep->next_in_bucket = hash_table[index];
    hash_table[index] = mep;
} /* hash_insert */

/*  This function takes an address, "pointer", as argument, and returns
 *  a pointer to the malloc entry describing the allocated area that
 *  "pointer" points to - or NULL, if there is no such entry.
 */
static struct malloc_entry *hash_lookup(pointer)
  char *pointer;
{
    int index = HASH(pointer);
    struct malloc_entry *mep = hash_table[index];

    while (mep != NULL)
	if (mep->the_area == pointer)
	    return mep;
	else
	    mep = mep->next_in_bucket;
    return NULL;
} /* hash_lookup */

/*---------------------------------------------------------------------------*/

static struct malloc_entry *first_malloc = NULL, *last_malloc = NULL;

/*  Create a new, empty, malloc_entry struct,
 *  put it last in the linked list, and add it to the hash table.
 */
static struct malloc_entry *new_malloc_entry(the_area)
  char *the_area;
{
    struct malloc_entry *mep;

    mep = (struct malloc_entry *)real_safe_malloc(sizeof(struct malloc_entry));
    mep->entry_number = nr_entries++;
    mep->the_area = the_area;
    mep->freed_file_name = NULL;
    mep->freed_line_number = 0;
    mep->checksum = -1;
    mep->realloced_to = NULL;
    mep->realloced_from = NULL;

    mep->next_in_sequence = NULL;
    if (first_malloc == NULL) {
	first_malloc = mep;
	last_malloc = mep;
    }
    else {
	last_malloc->next_in_sequence = mep;
	last_malloc = mep;
    }

    hash_insert(mep);
    return mep;
} /* new_malloc_entry */

/*---------------------------------------------------------------------------*/

static void set_guard_bytes(the_area, the_size)
  register char *the_area;
  register SIZE_T the_size;
{
    the_area[-8] = GUARD_BYTE;
    the_area[-7] = GUARD_BYTE;
    the_area[-6] = GUARD_BYTE;
    the_area[-5] = GUARD_BYTE;
    the_area[-4] = GUARD_BYTE;
    the_area[-3] = GUARD_BYTE;
    the_area[-2] = GUARD_BYTE;
    the_area[-1] = GUARD_BYTE;
    the_area[the_size] = GUARD_BYTE;
    the_area[the_size + 1] = GUARD_BYTE;
    the_area[the_size + 2] = GUARD_BYTE;
    the_area[the_size + 3] = GUARD_BYTE;
    the_area[the_size + 4] = GUARD_BYTE;
    the_area[the_size + 5] = GUARD_BYTE;
    the_area[the_size + 6] = GUARD_BYTE;
    the_area[the_size + 7] = GUARD_BYTE;
} /* set_guard_bytes */

/*---------------------------------------------------------------------------*/

static void print_area(file, pointer, nr_bytes)
  FILE *file;
  register unsigned char *pointer;
  SIZE_T nr_bytes;
{
    int i;
    SIZE_T n;

    if (nr_bytes < dm_max_print_length)
	n = nr_bytes;
    else
	n = dm_max_print_length;

    fputc('\'', file);
    for (i = 0; i < n; ++i) {
        int this_char;

        this_char = pointer[i];
        if(isprint(this_char))
	    fputc(this_char, file);
        else if (dm_print_codes)
	    fprintf(file, "(%d)", this_char); /* not printable */
	else
	    fputc(dm_print_char, file); /* not printable */
    } /* for */

    fputc('\'', file);
    if (nr_bytes > dm_max_print_length)
	fprintf(file, "...");
} /* print_area */

/*---------------------------------------------------------------------------*/

/*  This function prints a description of a malloc entry, pointed to by
 *  "mep", on the file "file". A checksum is calculated and compared
 *  to the stored checksum, but otherwise no checks are done, just printing.
 */
static void report_entry(file, mep)
  FILE *file;
  register struct malloc_entry *mep;
{

    fprintf(file, "*** Reporting malloc entry number %ld:\n", mep->entry_number);
    if (mep->remark)
	fprintf(file, "    Remark: %s\n", mep->remark);
    if (mep->realloced_from) {
	fprintf(file, "    Reallocated from address %p (= hex %lx) (entry number %ld) with size %d\n",
		mep->realloced_from->the_area,
		(unsigned long int)mep->realloced_from->the_area,
		mep->realloced_from->entry_number,
		mep->realloced_from->the_size);
	if (mep->realloced_from->freed_file_name)
	    fprintf(file, "        at line number %d in file '%s'.\n",
		    mep->realloced_from->freed_line_number,
		    mep->realloced_from->freed_file_name);
	else
	    fprintf(file, "        at an unknown place, not using debug_realloc.\n");
    }
    else {
	if (mep->file_name)
	    fprintf(file, "    Allocated at line number %d in file '%s'.\n",
		    mep->line_number, mep->file_name);
	else
	    fprintf(file, "    Allocated at an unknown place, not using the debug_malloc calls.\n");
    }

    if (mep->status == used) {
	fprintf(file, "    Status: USED\n");
    }
    else if (mep->status == freed) {
	fprintf(file, "    Status: FREED\n");
	if (mep->freed_file_name)
	    fprintf(file, "    Free'd at line number %d in file '%s'.\n",
		    mep->freed_line_number, mep->freed_file_name);
	else
	    fprintf(file, "    Free'd at an unknown place, not using the debug_malloc calls.\n");
    }
    else {
	fprintf(file, "    Status: REALLOCED\n");
	fprintf(file, "    Reallocated to address %p (= hex %lx) (entry number %ld) with size %d\n",
		mep->realloced_to->the_area,
		(unsigned long int)mep->realloced_to->the_area,
		mep->realloced_to->entry_number,
		mep->realloced_to->the_size);
	if (mep->freed_file_name)
	    fprintf(file, "        at line number %d in file '%s'.\n",
		    mep->freed_line_number, mep->freed_file_name);
	else
	    fprintf(file, "        at an unknown place, not using debug_realloc.\n");
    }

    fprintf(file, "    Address of malloced area: %p (= hex %lx)\n",
	    mep->the_area, (unsigned long int)mep->the_area);
    fprintf(file, "    (Recorded) size of malloced area: %d\n", mep->the_size);
    if (mep->status != used && mep->checksum != calculate_checksum(mep->the_area, mep->the_size)) {
	fprintf(file, "    Checksum error: stored = %d, calculated = %d\n", mep->checksum,
		calculate_checksum(mep->the_area, mep->the_size));
    }
    fprintf(file, "    Preceding guard bytes (dec): %d %d %d %d %d %d %d %d\n",
	    mep->the_area[-8], mep->the_area[-7], mep->the_area[-6], mep->the_area[-5],
	    mep->the_area[-4], mep->the_area[-3], mep->the_area[-2], mep->the_area[-1]);
    fprintf(file, "    Trailing guard bytes (dec): %d %d %d %d %d %d %d %d\n",
	    mep->the_area[mep->the_size],
	    mep->the_area[mep->the_size + 1],
	    mep->the_area[mep->the_size + 2],
	    mep->the_area[mep->the_size + 3],
	    mep->the_area[mep->the_size + 4],
	    mep->the_area[mep->the_size + 5],
	    mep->the_area[mep->the_size + 6],
	    mep->the_area[mep->the_size + 7]);
    fprintf(file, "    Contents: ");
    print_area(file, mep->the_area, mep->the_size);
    fprintf(file, "\n");
} /* report_entry */

/*---------------------------------------------------------------------------*/

/*  This function performs some checks on the malloc entry pointed to by
 *  "mep"; 1) a checksum is calculated and compared to the stored checksum,
 *  and 2) the "guard bytes" are examined.
 *  If an error was found, "report_entry" is called.
 *  The function returns the number of problems it found.
 */
static int check_one_entry(file, mep, fun_name, file_name, line_nr)
  FILE *file;
  struct malloc_entry *mep;
  char *fun_name, *file_name;
  int line_nr;
{
    int nr_warnings_found = 0;
    
    if (   (mep->status == freed || mep->status == realloced)
	&& mep->checksum != calculate_checksum(mep->the_area, mep->the_size)) {
	if (nr_warnings_found == 0)
	    fprintf(file, "\n");
	warning(file, fun_name, file_name, line_nr,
		"Unused area (address %p (= hex %lx), entry number %ld) has changed.",
		mep->the_area, (unsigned long int)mep->the_area, mep->entry_number);
	++nr_warnings_found;
    }
    if (   mep->the_area[-8] != GUARD_BYTE
	|| mep->the_area[-7] != GUARD_BYTE
	|| mep->the_area[-6] != GUARD_BYTE
	|| mep->the_area[-5] != GUARD_BYTE
	|| mep->the_area[-4] != GUARD_BYTE
	|| mep->the_area[-3] != GUARD_BYTE
	|| mep->the_area[-2] != GUARD_BYTE
	|| mep->the_area[-1] != GUARD_BYTE) {
	if (nr_warnings_found == 0)
	    fprintf(file, "\n");
	warning(file, fun_name, file_name, line_nr,
		"Preceding guard bytes damaged in entry number %ld.",
		mep->entry_number);
	++nr_warnings_found;
    }
    if (   mep->the_area[mep->the_size] != GUARD_BYTE
	|| mep->the_area[mep->the_size + 1] != GUARD_BYTE
	|| mep->the_area[mep->the_size + 2] != GUARD_BYTE
	|| mep->the_area[mep->the_size + 3] != GUARD_BYTE
	|| mep->the_area[mep->the_size + 4] != GUARD_BYTE
	|| mep->the_area[mep->the_size + 5] != GUARD_BYTE
	|| mep->the_area[mep->the_size + 6] != GUARD_BYTE
	|| mep->the_area[mep->the_size + 7] != GUARD_BYTE) {
	if (nr_warnings_found == 0)
	    fprintf(file, "\n");
	warning(file, fun_name, file_name, line_nr,
		"Trailing guard bytes damaged in entry number %ld.",
		mep->entry_number);
	++nr_warnings_found;
    }
    if (nr_warnings_found)
	report_entry(file, mep);
    if (nr_warnings_found && dm_repair) {
	/* Ok, maybe the error wasn't in the guard bytes, but then this is just a no-op. */
	set_guard_bytes(mep->the_area, mep->the_size);
	/* The same goes for the checksum - at worst, this is just a waste of time. */
	mep->checksum = calculate_checksum(mep->the_area, mep->the_size);
    }
    return nr_warnings_found;
} /* check_one_entry */

/*---------------------------------------------------------------------------*/

static int check_all_entries(file, fun_name, file_name, line_nr)
  FILE *file;
  char *fun_name, *file_name;
  int line_nr;
{
    register struct malloc_entry *mep = first_malloc;
    int nr_bad_entries = 0;

    while (mep) {
	if (check_one_entry(file, mep, fun_name, file_name, line_nr))
	    ++nr_bad_entries;
	mep = mep->next_in_sequence;
    } /* while */
    return nr_bad_entries;
} /* check_all_entries */

/*---------------------------------------------------------------------------*/

/*  If it was possible to allocate the requested number of bytes,
 *  a malloc entry will be created and stored in the table,
 *  and the function will return a pointer to the allocated area.
 *  If it wasn't possible, no such entry is created,
 *  and the function returns NULL.
 */
static char *do_allocate(size, file_name, line_number)
  SIZE_T size;
  int line_number;
  char *file_name;
{
    register struct malloc_entry *mep;
    register char *cp;

    cp = real_malloc(size + 16);
    if (cp != NULL) {
	cp += 8;
	set_guard_bytes(cp, size);
	mep = new_malloc_entry(cp);
	mep->file_name = file_name;
	mep->line_number = line_number;
	mep->the_size = size;
	mep->status = used;
	mep->remark = dm_global_remark;
	return mep->the_area;
    }
    else
	return NULL;
} /* do_allocate */

/*---------------------------------------------------------------------------*/

char *debug_malloc(size, file_name, line_number)
  SIZE_T size;
  char *file_name;
  int line_number;
{
    char *cp;

    ++malloc_calls;

    if (dm_insane)
	check_all_entries(dm_error_stream, "malloc", file_name, line_number);

    cp = do_allocate(size, file_name, line_number);
    if (cp == NULL)
        malloc_failed("malloc: Couldn't allocate %d bytes.", size);
    else if (dm_fill_new)
	set_area(cp, size, FILL_NEW_BYTE);
    return cp;	/* Might be NULL! */
} /* debug_malloc */

char *debug_calloc(nr_elements, element_size, file_name, line_number)
  unsigned int nr_elements;
  SIZE_T element_size;
  char *file_name;
  int line_number;
{
    register char *cp;
    int i;
    unsigned long int lsize;
    SIZE_T size;

    ++calloc_calls;

    if (dm_insane)
	check_all_entries(dm_error_stream, "calloc", file_name, line_number);

    lsize = (unsigned long int)nr_elements * (unsigned long int)element_size;
    if (lsize > MAX_MALLOC_SIZE) {
	malloc_failed("calloc: Couldn't allocate %ld bytes.", lsize);
	return NULL;
    }

    size = (int)lsize;

    cp = do_allocate(size, file_name, line_number);
    if (cp == NULL)
        malloc_failed("calloc: Couldn't allocate %d bytes.", size);
    else if (cp != NULL)
	set_area(cp, size, 0);
    return cp;
} /* debug_calloc */

void debug_free(pointer, file_name, line_number)
  char *pointer;
  char *file_name;
  int line_number;
{
    register struct malloc_entry *mep;

    ++free_calls;

    mep = hash_lookup(pointer);
    if (mep == NULL) {
	fprintf(dm_error_stream, "\n");
	warning(dm_error_stream, "free", file_name, line_number,
		"Tried to free an unallocated area (address %p = hex %lx).",
		pointer, (unsigned long int)pointer);
	return;
    }
    if (mep->status == freed) {
	fprintf(dm_error_stream, "\n");
	warning(dm_error_stream, "free", file_name, line_number,
		"Tried to free an already freed area (address %p = hex %lx, entry number %ld).",
		pointer, (unsigned long int)pointer, mep->entry_number);
	report_entry(dm_error_stream, mep);
	return;
    }
    if (mep->status == realloced) {
	fprintf(dm_error_stream, "\n");
	warning(dm_error_stream, "free", file_name, line_number,
		"Tried to free a realloced area (address %p = hex %lx, entry number %ld).",
		pointer, (unsigned long int)pointer, mep->entry_number);
	report_entry(dm_error_stream, mep);
	return;
    }
    mep->status = freed;
    mep->freed_file_name = file_name;
    mep->freed_line_number = line_number;
    if (dm_fill_freed)
	set_area(mep->the_area, mep->the_size, FILL_FREED_BYTE);
    mep->checksum = calculate_checksum(mep->the_area, mep->the_size);

    if (dm_insane)
	check_all_entries(dm_error_stream, "free", file_name, line_number);
    else
	check_one_entry(dm_error_stream, mep, "free", file_name, line_number);
} /* debug_free */

char *debug_realloc(pointer, size, file_name, line_number)
  char *pointer;
  int line_number;
  SIZE_T size;
  char *file_name;
{
    register struct malloc_entry *oldmep, *newmep;
    char *new_area;

    ++realloc_calls;

    oldmep = hash_lookup(pointer);
    if (oldmep == NULL) {
	fprintf(dm_error_stream, "\n");
	warning(dm_error_stream, "realloc", file_name, line_number,
		"Tried to realloc an unallocated area (address %p = hex %lx).",
		pointer, (unsigned long int)pointer);
	return NULL;
    }
    if (oldmep->status == freed) {
	fprintf(dm_error_stream, "\n");
	warning(dm_error_stream, "realloc", file_name, line_number,
		"Tried to realloc a freed area (address %p = hex %lx, entry number %ld).",
		pointer, (unsigned long int)pointer, oldmep->entry_number);
	report_entry(dm_error_stream, oldmep);
	return NULL;
    }
    if (oldmep->status == realloced) {
	fprintf(dm_error_stream, "\n");
	warning(dm_error_stream, "realloc", file_name, line_number,
		"Tried to realloc an already realloced area (address %p = hex %lx, entry number %ld).",
		pointer, (unsigned long int)pointer, oldmep->entry_number);
	report_entry(dm_error_stream, oldmep);
	return NULL;
    }

    if (dm_insane)
	check_all_entries(dm_error_stream, "realloc", file_name, line_number);
    else
	check_one_entry(dm_error_stream, oldmep, "realloc", file_name, line_number);

    new_area = do_allocate(size, file_name, line_number);
    if (new_area == NULL) {
	malloc_failed("realloc: Couldn't allocate %d bytes.", size);
	return NULL;
    }

    newmep = last_malloc;	/* The one just allocated. */
    oldmep->status = realloced;
    oldmep->freed_file_name = file_name;
    oldmep->freed_line_number = line_number;
    oldmep->realloced_to = newmep;
    newmep->realloced_from = oldmep;

    if (oldmep->the_size < size) {
	copy_bytes(oldmep->the_area, newmep->the_area, oldmep->the_size);
	if (dm_fill_new)
	    set_area(newmep->the_area + oldmep->the_size, size - oldmep->the_size, FILL_NEW_BYTE);
    }
    else
	copy_bytes(oldmep->the_area, newmep->the_area, size);

    if (dm_fill_freed)
	set_area(oldmep->the_area, oldmep->the_size, FILL_FREED_BYTE);
    oldmep->checksum = calculate_checksum(oldmep->the_area, oldmep->the_size);

    return new_area;
} /* debug_realloc */

/*---------------------------------------------------------------------------*/

/*  This function prints the report about the malloc entry pointed to by
 *  the pointer "pointer", if such an area exists.
 */
int dm_whatis_pointer(pointer)
  char *pointer;
{
    register struct malloc_entry *mep;
    int check_result;

    mep = hash_lookup(pointer);
    if (mep == NULL) {
	printf("dm_whatis_pointer (debug_malloc): Never heard of such a pointer (%p = hex %lx)\n",
	       pointer, (unsigned long int)pointer);
	return -1;
    }
    check_result = check_one_entry(stdout, mep, "no function", "no file", 0);
    report_entry(stdout, mep);
    return check_result;
} /* dm_whatis_pointer */

/*---------------------------------------------------------------------------*/

int dm_check_all()
{
    int nr_bad_entries;

    nr_bad_entries = check_all_entries(dm_error_stream, (char *)NULL, (char *)NULL, 0);
    if (nr_bad_entries > 0)
	fprintf(dm_error_stream, "\n*** dm_check_all: found %d malloc entries with errors\n",
		nr_bad_entries);
    return nr_bad_entries;
} /* dm_check_all */

/*---------------------------------------------------------------------------*/

/* Just used for my own debugging of debug_malloc itself */

static void print_hash_statistics(outfile)
  FILE *outfile;
{
    int hash_value, max_chain_hash;
    struct malloc_entry *bucketp;
    long int chain_length,	/* The length of the chain in one bucket */
             nr_1chains,	/* Number of chains >= 1 elements */
             nr_2chains,	/* Number of chains >= 2 elements */
    	     sum_chain_length,
             max_chain_length;

    nr_1chains = 0;
    nr_2chains = 0;
    sum_chain_length = 0;
    max_chain_length = 0;
    max_chain_hash = -1;
    for (hash_value  = 0; hash_value < HASH_TABLE_SIZE; ++hash_value) {
	bucketp = hash_table[hash_value];
	chain_length = 0;
	while (bucketp != NULL) {
	    ++chain_length;
	    bucketp = bucketp->next_in_bucket;
	} /* while elements left in this chain */
	if (chain_length > 0)
	    ++nr_1chains;
	if (chain_length > 1)
	    ++nr_2chains;
	if (chain_length > max_chain_length) {
	    max_chain_length = chain_length;
	    max_chain_hash = hash_value;
	}
	sum_chain_length += chain_length;
    } /* for all possible hash values */

    fprintf(outfile, "Statistics for debug_malloc's internal hash table:\n");
    fprintf(outfile, "--------------------------------------------------\n");
    fprintf(outfile, "    Hash size: %d\n", HASH_TABLE_SIZE);
    fprintf(outfile, "    Number of entries: %ld (load factor: %d %%)\n",
	    sum_chain_length, (100 * sum_chain_length) / HASH_TABLE_SIZE);

    if (sum_chain_length == 0)
	return;

    if (max_chain_hash != -1)
	fprintf(outfile, "    Longest chain: %ld (for hash value %d)\n",
		max_chain_length, max_chain_hash);
    else
	fprintf(outfile, "    Longest chain: -\n");

    fprintf(outfile, "    Number of empty buckets: %ld (%d %%)\n",
	    HASH_TABLE_SIZE - nr_1chains,
	    (100 * (HASH_TABLE_SIZE - nr_1chains)) / HASH_TABLE_SIZE);
    fprintf(outfile, "    Number of chains longer than 0: %ld (%d %%)\n",
	    nr_1chains, (100 * nr_1chains) / HASH_TABLE_SIZE);
    fprintf(outfile, "    Number of chains longer than 1: %ld (%d %%)\n",
	    nr_2chains, (100 * nr_2chains) / HASH_TABLE_SIZE);
    fprintf(outfile, "    Average chain length: %d.%d\n",
	    (int)(sum_chain_length / HASH_TABLE_SIZE),
	    (int)(((100 * sum_chain_length) / HASH_TABLE_SIZE) % 100));
    fprintf(outfile, "    Average length of chains longer than 0: %d.%d\n",
	    (int)(sum_chain_length / nr_1chains),
	    (int)(((100 * sum_chain_length) / nr_1chains) % 100));
} /* print_hash_statistics */

/*---------------------------------------------------------------------------*/

/*  This function goes through all the malloc entries, and prints a report
 *  for each one. Then, all entries are checked for errors.
 *  The messages are printed on a file with its name in "dm_dump_file_name".
 */
static void internal_dump(full_flag)
  int full_flag;
{
    FILE *dump_file;
    long int nr_used = 0, nr_freed = 0, nr_realloced = 0, used_bytes = 0, total_bytes = 0;
    int nr_bad_entries;
    register struct malloc_entry *mep;
    long int the_time;
    int check_result;

    fprintf(dm_error_stream,
	    "\n*** The debug_malloc package will now dump its %sstatus on '%s'.\n",
	    (full_flag ? "FULL " : ""),
	    dm_dump_file_name);

    dump_file = fopen(dm_dump_file_name, "w");
    if (dump_file == NULL) {
	fprintf(dm_error_stream,
		"\n*** The debug_malloc package couldn't open '%s' for writing.\n",
		dm_dump_file_name);
	return;
    }

    mep = first_malloc;
    while (mep) {
	if (mep->status == used) {
	    ++nr_used;
	    used_bytes += mep->the_size;
	}
	else if (mep->status == freed)
	    ++nr_freed;
	else if (mep->status == realloced)
	    ++nr_realloced;
	total_bytes += mep->the_size;
	mep = mep->next_in_sequence;
    } /* while */

    the_time = time((long int *)NULL);

    if (full_flag) {
	fprintf(dump_file, "FULL dump of malloc status (%24.24s):\n", ctime(&the_time));
	fprintf(dump_file, "------------------------------------------------------\n");
    }
    else {
	fprintf(dump_file, "Dump of malloc status (%24.24s):\n", ctime(&the_time));
	fprintf(dump_file, "-------------------------------------------------\n");
    }

    fprintf(dump_file, "Version: %s\n", dm_version);
    fprintf(dump_file, "    Compiled with: ");
#ifdef MSDOS
    fprintf(dump_file, "MSDOS defined, alignment factor = %d\n", ALIGNMENT_FACTOR);
#else
    fprintf(dump_file, "MSDOS not defined, alignment factor = %d\n", ALIGNMENT_FACTOR);
#endif
#ifdef USE_STANDARD_MALLOC
    fprintf(dump_file, "    Compiled to call standard malloc.\n");
#else
#   ifdef MSDOS
    fprintf(dump_file, "    Compiled to use its %d own buffers of each %ld bytes.\n",
	    NR_MSDOS_MALLOC_BUFFERS, (long)MSDOS_MALLOC_BUFFER_SIZE);
#   else
    fprintf(dump_file, "    Compiled to use its own buffer with size %ld bytes.\n",
	    (long)MALLOC_BUFFER_SIZE);
#   endif
#endif

    fprintf(dump_file, "    Total number of malloc entries: %ld\n", nr_entries);
    fprintf(dump_file, "    Number of malloc entries still in use: %ld\n", nr_used);
    fprintf(dump_file, "    Number of freed malloc entries: %ld\n", nr_freed);
    fprintf(dump_file, "    Number of reallocated malloc entries: %ld\n", nr_realloced);
    fprintf(dump_file, "    Total number of bytes allocated: %ld\n", total_bytes);
    fprintf(dump_file, "    Number of bytes still in use: %ld\n", used_bytes);
    fprintf(dump_file, "    Number of calls to malloc: %ld, calloc: %ld, realloc: %ld, free: %ld\n",
	    malloc_calls, calloc_calls, realloc_calls, free_calls);
    fprintf(dump_file, "    dm_dump_file_name = '%s'\n", dm_dump_file_name);
    fprintf(dump_file, "    dm_insane = %d, dm_repair = %d\n",
	    dm_insane, dm_repair);
    fprintf(dump_file, "    dm_print_codes = %d, dm_print_char = '%c', dm_max_print_length = %d\n",
	    dm_print_codes, dm_print_char, dm_max_print_length);
    fprintf(dump_file, "\n");

    print_hash_statistics(dump_file);
    fprintf(dump_file, "\n");

    if (full_flag) {
	fprintf(dump_file, "FULL listing (and checking) of ALL malloc entries:\n");
	fprintf(dump_file, "--------------------------------------------------\n");
    }
    else {
	fprintf(dump_file, "Listing all malloc entries still in use:\n");
	fprintf(dump_file, "----------------------------------------\n");
    }

    if (full_flag) {
	nr_bad_entries = 0;
	mep = first_malloc;
	while (mep) {
	    report_entry(dump_file, mep);
	    fprintf(dump_file, "\n");
	    check_result = check_one_entry(dump_file, mep, "dm_full_dump", "no file", 0);
	    if (check_result)
		++nr_bad_entries;
	    mep = mep->next_in_sequence;
	} /* while */
    }
    else {
	mep = first_malloc;
	while (mep) {
	    if (mep->status == used) {
		if (mep->file_name)
		    fprintf(dump_file, "%ld: file '%s' line %d: %p (= hex %lx) (%d bytes): \n",
			    mep->entry_number, mep->file_name, mep->line_number,
			    mep->the_area, (unsigned long int)mep->the_area, mep->the_size);
		else
		    fprintf(dump_file, "%ld: allocated at an unknown place: %p (= hex %lx) (%d bytes): \n",
			    mep->entry_number, mep->the_area,
			    (unsigned long int)mep->the_area, mep->the_size);
		if (mep->remark)
		    fprintf(dump_file, "    Remark: %s\n", mep->remark);
		print_area(dump_file, mep->the_area, mep->the_size);
		fprintf(dump_file, "\n");
	    }
	    mep = mep->next_in_sequence;
	} /* while */
    } /* if not full_flag */

    if (!full_flag) {
	fprintf(dump_file, "\n");
	fprintf(dump_file, "Checking all malloc entries...\n");
	fprintf(dump_file, "------------------------------\n");
	nr_bad_entries = check_all_entries(dump_file, (char *)NULL, (char *)NULL, 0);
    }

    fprintf(dump_file, "\n");
    fprintf(dump_file, "There were %d malloc entries with errors in them.\n", nr_bad_entries);
    fprintf(dump_file, "Finished dumping the malloc status. Have fun debugging!\n");

    fclose(dump_file);
    fprintf(dm_error_stream,
	    "    The debug_malloc package has now finished dumping its %sstatus.\n",
	    (full_flag ? "FULL " : ""));
    fprintf(dm_error_stream, "    There were %d bad entries.\n", nr_bad_entries);
} /* internal_dump */

/*---------------------------------------------------------------------------*/

/*  "dm_dump" goes through all the malloc entries that are used,
 *  and prints a short report for each one.
 *  Then, all entries are checked for errors.
 *  The messages are printed on a file with its name in "dm_dump_file_name".
 *  "dm_full_dump" prints a long report for all malloc entries,
 *  both used and unused, and at the same time checks them for errors.
 */
void dm_dump() { internal_dump(0); }
void dm_full_dump() { internal_dump(1); }

/*---------------------------------------------------------------------------*/

/*  This is useful if some parts of your program, or the libraries it uses,
 *  have not been compiled to call debug_malloc, debug_realloc etc.
 */

#ifndef USE_STANDARD_MALLOC

char *malloc(s)
  SIZE_T s;
{
    return debug_malloc(s, (char *)NULL, 0);
} /* malloc */

char *calloc(n, s)
  int n;
  SIZE_T s;
{
    return debug_calloc(n, s, (char *)NULL, 0);
} /* calloc */

char *realloc(p, s)
  char *p;
  SIZE_T s;
{
    return debug_realloc(p, s, (char *)NULL, 0);
} /* realloc */

void free(p)
  char *p;
{
    debug_free(p, (char *)NULL, 0);
} /* free */

#endif

/*---------------------------------------------------------------------------*/

/*  bottom_malloc works as the standard malloc, except that it
 *  allocates memory in a big buffer instead of dynamically.
 *
 *  Note that there is no bottom_calloc, bottom_realloc or bottom_free,
 *  since they are never used by the SM/DM package!
 *  Consequently, bottom_malloc can be implemented in a very simple
 *  way, which is how it was done - so if you need a REAL malloc
 *  package, "bottom_malloc" will not work.
 *
 */

#ifndef USE_STANDARD_MALLOC

#ifdef MSDOS

/* This code is a simple replacement for malloc for MSDOS */

/* static char *malloc_buffers [ MSDOS_MALLOC_BUFFER_SIZE ] [ NR_MSDOS_MALLOC_BUFFERS ]; */

static char malloc_buffer0 [ MSDOS_MALLOC_BUFFER_SIZE ];

static char *malloc_buffers [ NR_MSDOS_MALLOC_BUFFERS ] = {
    malloc_buffer0
};

static int used_bytes [ NR_MSDOS_MALLOC_BUFFERS ];
static int used_buffers = 0;

static char *bottom_malloc(size)
  SIZE_T size;
{
    char *area_start;
    int i;

    /* All allocated areas should start on even ALIGNMENT_FACTOR-byte addresses: */
    if (size % ALIGNMENT_FACTOR != 0)
	size += ALIGNMENT_FACTOR - size % ALIGNMENT_FACTOR;

    for (i = 0; i < used_buffers; ++i) {
	if (used_bytes[i] + size <= MSDOS_MALLOC_BUFFER_SIZE) {
	    area_start = malloc_buffers[i] + used_bytes[i];
	    used_bytes[i] += size;
	    return area_start;
	}
    }

    if (used_buffers >= NR_MSDOS_MALLOC_BUFFERS || size > MSDOS_MALLOC_BUFFER_SIZE)
	return NULL;

    area_start = malloc_buffers[used_buffers] + used_bytes[used_buffers];
    used_bytes[used_buffers++] += size;

    return area_start;
} /* bottom_malloc */

#else

/*  This code is a simple replacement for malloc for systems
 *  without those stupid MSDOS 64K and 640K limits.
 */

static char malloc_buffer [ MALLOC_BUFFER_SIZE ];
static int used_bytes = 0;

static char *bottom_malloc(size)
  SIZE_T size;
{
    char *area_start;

    /* All allocated areas should start on even ALIGNMENT_FACTOR-byte addresses: */
    if (size % ALIGNMENT_FACTOR != 0)
	size += ALIGNMENT_FACTOR - size % ALIGNMENT_FACTOR;
    if (used_bytes + size > MALLOC_BUFFER_SIZE)
	return NULL;

    area_start = malloc_buffer + used_bytes;
    used_bytes += size;

    return area_start;
} /* bottom_malloc */

#endif

#endif
