/*  safe_malloc.c -- a "checking malloc" package with a debug option
 *  Thomas Padron-McCarthy (padrone@lysator.liu.se), 1991
 *  Latest change of this file: September 24, 1992
 *
 *  This module, "safe_malloc", is part of the safe_malloc/debug_malloc
 *  (SM/DM) package for malloc debugging.
 *
 *  These functions are used as an interface to the standard
 *  malloc calls. If a malloc (or realloc) call fails,
 *  the function "malloc_failed" -- which should be defined by the
 *  application -- is called. If that function returns,
 *  NULL will then be returned from the malloc (or realloc).
 *  Also, realloc(NULL, ...), free(NULL), malloc(0, ...), calloc(0, ...)
 *  and realloc(NULL, ...) are guaranteed to give sensible results on
 *  (I hope) all systems.
 *
 *  The smart thing is not only that you don't have to check
 *  every malloc, but also that you can set the option DEBUG_MALLOC,
 *  and then use debug_malloc.c instead of this file, safe_malloc.c!
 *
 */

#include <malloc.h>

/* Don't #include "safe_malloc.h" here - it could re-define malloc etc! */

#include "smdm_config.h"

extern void malloc_failed();

/*---------------------------------------------------------------------------*/

char *safe_malloc(s)
  SIZE_T s;
{
    char *temp;
    
    temp = malloc(s ? s : 1);
    if (temp == NULL)
	malloc_failed("malloc: Couldn't allocate %d bytes.", s);
    return temp;
} /* safe_malloc */

char *safe_calloc(n, s)
  unsigned int n;
  SIZE_T s;
{
    char *temp;
    
    temp = calloc(n ? n : 1, s ? s : 1);
    if (temp == NULL)
	malloc_failed("calloc: Couldn't allocate %d bytes.", n * s);
    return temp;
} /* safe_calloc */

char *safe_realloc(p, s)
  char *p;
  SIZE_T s;
{
    char *temp;
    
    if (p == NULL)
	temp = malloc(s ? s : 1);
    else
	temp = realloc(p, s ? s : 1);
    if (temp == NULL)
	malloc_failed("realloc: Couldn't allocate %d bytes.", s);
    return temp;
} /* safe_realloc */

void safe_free(p)
  char *p;
{
    if (p != NULL)
	free(p);
} /* safe_free */
