/*  test_malloc.c
 *	-- a program to test malloc debugging
 *  Thomas Padron-McCarthy (padrone@lysator.liu.se), 1991, 1992
 *  Latest change of this file: Oct 3, 1992
 *
 *  This program, "test_malloc", is part of the safe_malloc/debug_malloc
 *  (SM/DM) package for malloc debugging.
 *
 */

#include <stdio.h>
#include <strings.h>
#include "safe_malloc.h"

#define ASSERT(what)       		\
        {                               \
                if (!(what))            \
                        test_malloc_failed("ASSERT failed in \"%s\" line %d: %s",	\
					   __FILE__, __LINE__, "what");                 \
        }

/* How many times to run through - one or more: */
#define NR_TESTS	1	
/* The number of calls in each test block - should be at least 18: */
#define BLOCK_SIZE	40
/* Should be at least 18, too: */
#define MAX_BYTES	170
/* We need to allocate a rather big area too: */
#define EXTRA_MAX_BYTES	1703


static char *rem_p;

/*---------------------------------------------------------------------------*/

extern int errno;
extern char *sys_errlist[];
extern int sys_nerr;

/*  This is NOT the "malloc_failed" function that is called from
 *  the SMDM package when a malloc call fails!
 *  This function is called when an ASSERT fails in this file!
 */

static void test_malloc_failed(fmt, a1, a2, a3, a4, a5)
  char *fmt, *a1, *a2, *a3, *a4, *a5;
{
    fflush(stdout);
    fflush(stderr);
    fprintf(stderr, "\n");
    fprintf(stderr, "\007");
    fprintf(stderr, "FATAL ERROR: ");
    fprintf(stderr, fmt, a1, a2, a3, a4, a5);
    fprintf(stderr, "\n");
    fprintf(stderr, "    Unix error number: %d (%s).\n",
            errno, sys_errlist[errno]);
    fprintf(stderr, "    Goodbye. I am sorry that this happened. Exiting.\n");
    fflush(stderr);
    exit(1);
} /* test_malloc_failed */

/*---------------------------------------------------------------------------*/

static int random_bytes()
{
    int s;
    
    s = rand() % MAX_BYTES;
    if (s == 17)
	s = EXTRA_MAX_BYTES;
    return s;
} /* random_bytes */

/*---------------------------------------------------------------------------*/

static void test_calloc() {
    char *cp;
    int i, n, n1, n2;

    fflush(stdout);
    fflush(stderr);
    printf("\nTesting: calloc...\n");
    fflush(stdout);
    fflush(stderr);
    n = random_bytes();
    cp = safe_calloc(1, n);
    for (i = 0; i < n; ++i)
	ASSERT(cp[i] == 0);
    safe_free(cp);

    n = random_bytes();
    cp = safe_calloc(n, 1);
    for (i = 0; i < n; ++i)
	ASSERT(cp[i] == 0);
    safe_free(cp);

    n1 = 5 + rand() % 10;
    n2 = random_bytes();
    cp = safe_calloc(n1, n2);
    for (i = 0; i < n1 * n2; ++i)
	ASSERT(cp[i] == 0);
    safe_free(cp);
} /* test_calloc */

static void do_test(tnr)
  int tnr;
{
    char *block[ BLOCK_SIZE ], *block2[ BLOCK_SIZE ];
    char *cp;
    int pnr, i, s;
    
    printf("Testing: Test number %d...\n", tnr);
    printf("Testing: mallocs in test number %d...\n", tnr);
    fflush(stdout);
    fflush(stderr);
    for (pnr = 0; pnr < BLOCK_SIZE; ++pnr) {
	s = random_bytes();
	block[pnr] = cp = safe_malloc(s);
	for (i = 0; i < s; ++i)
	    cp[i] = 'a' + i % 26;
    }
    printf("Testing: reallocs in test number %d...\n", tnr);
    block[10][-3] = 'X';
    fflush(stdout);
    fflush(stderr);
    printf("----------------------------------------\n");
    printf("One warning now in safe_realloc, in test %d, pointer %lx:\n",
	   tnr, (unsigned long int)block[10]);
    printf("----------------------------------------\n");
    fflush(stdout);
    fflush(stderr);
    for (pnr = 0; pnr < BLOCK_SIZE; ++pnr) {
	s = random_bytes();
	block2[pnr] = cp = safe_realloc(block[pnr], s);
	for (i = 0; i < s; ++i)
	    cp[i] = '0' + i % 10;
    }
    printf("Testing: frees in test number %d...\n", tnr);
    if (strlen(block[11]) < 4)
	printf("**** Ouch!!! ***\n");
    block[11][3] = 'Y';
    fflush(stdout);
    fflush(stderr);
    printf("----------------------------------------\n");
    printf("If dm_insane (which is %d): one warning now in safe_free, in test %d, pointer %lx:\n",
#ifdef DEBUG_MALLOC
	   dm_insane,
#else
	   0,
#endif
	   tnr, (unsigned long int)block[11]);
    rem_p = block[11];
    printf("----------------------------------------\n");
    fflush(stdout);
    fflush(stderr);
    for (pnr = 0; pnr < BLOCK_SIZE; ++pnr)
	safe_free(block2[pnr]);
    fflush(stdout);
    fflush(stderr);
    printf("----------------------------------------\n");
    printf("Six warnings now, in test %d:\n", tnr);
    printf("----------------------------------------\n");
    fflush(stdout);
    fflush(stderr);
    safe_free((char *)NULL);
    safe_realloc((char *)NULL, 90);
    safe_free(block[0]);
    safe_free(block[17]);
    safe_free(block2[0]);
    safe_free(block2[17]);

    test_calloc();

} /* do_test */

int main()
{
    int tnr;

    printf("Welcome to test_malloc, a program to test the debug_malloc package!\n");
    printf("\n");

#ifndef DEBUG_MALLOC
    printf("This program will not work properly unless in debug mode\n");
    printf("(i. e. with DEBUG_MALLOC #defined).\n");
    printf("It compiles, but since it overwrites some data\n");
    printf("- things that the debug mode catches - it will probably crash\n");
    printf("or at last produce very strange results.\n");
    printf("Press RETURN to continue anyway!\n");
    getchar();
#endif

    printf("You will get some warnings from the debug malloc package,\n");
    printf("saying \"*** MALLOC WARNING\", followed by an error report.\n");
    printf("These are normal, since this program intentionally causes errors\n");
    printf("to test that the malloc package can catch them.\n");
    printf("You will be told which such warnings to expect,\n");
    printf("and if you DON'T see them, the debug malloc package did NOT work as intended.\n");
    printf("\n");
    printf("If you, on the other hand, see a message saying\n");
    printf("\"FATAL ERROR: ASSERT failed\", and some more text,\n");
    printf("after which the program terminates,\n");
    printf("the debug malloc package did not work as intended.\n");
    printf("\n");
    printf("Press RETURN to continue!\n");
    getchar();

#ifdef DEBUG_MALLOC
    dm_insane = 0;	/* You can change this to 1 to test "insane checking"! */
#endif

    srand(0);

    printf("Testing: Starting...\n");
    fflush(stdout);
    fflush(stderr);
    for (tnr = 1; tnr <= NR_TESTS; ++tnr)
	do_test(tnr);

/*
#ifdef DEBUG_MALLOC
    dm_dump();
    dm_dump_file_name = "malloc2.dmp";
    dm_full_dump();
#endif
*/

#ifdef DEBUG_MALLOC
    dm_insane = 1;
#endif

    {
	char *s;

	fflush(stdout);
	fflush(stderr);
	printf("----------------------------------------\n");
	printf("Note: If you didn't get a warning about pointer %lx earlier,\n",
	       (unsigned long int)rem_p);
	printf("it might come here:\n");
	printf("----------------------------------------\n");
	fflush(stdout);
	fflush(stderr);

	s = safe_malloc(8000);
	s[-2] = 14;
	s[8001] = 15;
	fflush(stdout);
	fflush(stderr);
	printf("----------------------------------------\n");
	printf("A warning will come here (pointer %lx):\n", (unsigned long int)s);
	printf("----------------------------------------\n");
	fflush(stdout);
	fflush(stderr);

	dm_check_all();
    }

#if 0
    /* If you want to do it an extra time. */
    fflush(stdout);
    fflush(stderr);
    printf("----------------------------------------\n");
    printf("*** Possibly LOTS of warnings now, not just those advertised...\n");
    printf("----------------------------------------\n");
    fflush(stdout);
    fflush(stderr);
    do_test(NR_TESTS + 1);
#endif

#ifdef DEBUG_MALLOC
    dm_dump();
    dm_dump_file_name = "malloc2.dmp";
    dm_full_dump();
#endif

    exit(0);
    /* return 0; -- Not reached */
} /* main */
